"""Tests for nodes from timeflux_dsp.nodes.spectral"""
import numpy as np
import pandas as pd
import pandas.util.testing as tm

from timeflux_dsp.nodes.spectral import Welch

# ------------------------------------------------------------------
# Create a signal for demonstration of Welch Periodogram .
# ------------------------------------------------------------------
# Here, we create three sinus with sampling rate 512 Hz, and of frequency 5, 15 and 50Hz,
# a dirac, a square, a centered dirac, mixed with white noise.
# We test if the peak frequency of the Welch node is correct.


# Prepare a reproducible example
np.random.seed(42)
rate = 512
n = rate * 10
m = 7
tm.K = m  # tm will generate 7 columns
original = tm.makeTimeDataFrame(n, freq="L").rename(
    columns={
        "A": "noise",
        "B": "dirac",
        "C": "centered dirac",
        "D": "square",
        "E": "sin5",
        "F": "sin15",
        "G": "sin50",
    }
)
original.index = original.index[0] + pd.to_timedelta(np.arange(n) / rate, unit="s")
t = (original.index - original.index[0]) / np.timedelta64(1, "s")
original["dirac"] = (original.index == original.index[0]).astype(float)
original["centered dirac"] = (original.index == original.index[n // 2]).astype(float)
original["square"] = (original.index <= original.index[rate]).astype(float)
original["sin5"] = 10 * np.sin(2 * np.pi * 5 * t) + original["noise"]
original["sin15"] = 10 * np.sin(2 * np.pi * 15 * t) + original["noise"]
original["sin50"] = 10 * np.sin(2 * np.pi * 50 * t) + original["noise"]


def test_welch():
    nperseg = 1024
    node = Welch(rate=rate, nperseg=nperseg, scaling="density")
    node.i.data = original
    node.update()

    expected_peaks = {
        "dirac": 0.0,
        "centered dirac": 1.0,
        "square": 0.5,
        "sin5": 5.0,
        "sin15": 15.0,
        "sin50": 50.0,
    }
    for space, expected_peak in expected_peaks.items():
        assert (
            node.o.data.loc[dict(space=space)]
            .frequency[node.o.data.loc[dict(space=space)].argmax("frequency")]
            .frequency.values[0]
            == expected_peak
        )



import pandas as pd
import numpy as np


class DummyData:
    """Generate dummy data."""

    def __init__(
        self,
        num_rows=1000,
        num_cols=5,
        rate=10,
        jitter=0.05,
        start_date="2018-01-01",
        seed=42,
        round=6,
    ):
        """
        Initialize the dataframe.

        Args:
            num_rows (int): Number of rows
            num_cols (int): Number of columns
            rate (float): Frequency, in Hertz
            jitter (float): Amount of jitter, relative to rate
            start_date (string): Start date
            seed (int): Seed for random number generation
            round (int): Number of decimals for random numbers
        """

        np.random.seed(seed)

        frequency = 1 / rate

        indices = pd.date_range(
            start=start_date, periods=num_rows, freq=pd.DateOffset(seconds=frequency)
        )

        jitter = frequency * jitter
        deltas = pd.to_timedelta(np.random.uniform(-jitter, jitter, num_rows), unit="s")
        indices = indices + deltas

        rows = np.random.rand(num_rows, num_cols).round(round)

        self._data = pd.DataFrame(rows, indices)
        self._cursor = 0

    def next(self, num_rows=10):
        """
        Get the next chunk of data.

        Args:
            num_rows (int): Number of rows to fetch
        """

        start = self._cursor
        stop = start + num_rows
        self._cursor += num_rows
        return self._data[start:stop]

    def reset(self):
        """
        Reset the cursor.
        """

        self._cursor = 0


class CustomData:
    """Generate dummy data."""

    def __init__(self, data):
        """
        Initialize the dataframe.

        Args:
            data (DataFrame): custom data to stream.
        """

        self._data = data
        self._cursor = 0

    def next(self, num_rows=10):
        """
        Get the next chunk of data.

        Args:
            num_rows (int): Number of rows to fetch
        """

        start = self._cursor
        stop = start + num_rows
        self._cursor += num_rows
        return self._data.iloc[start:stop]

    def reset(self):
        """
        Reset the cursor.
        """

        self._cursor = 0



"""Tests for iir and fir nodes"""

import numpy as np
import pandas as pd
import pandas.util.testing as tm
import pytest
from timeflux.helpers.testing import ReadData, Looper
from timeflux_dsp.nodes.filters import FIRFilter, IIRFilter


@pytest.fixture(scope="module")
def rate():
    return 50


# ------------------------------------------------------------------
# Create a signal for demonstration of IIR/FIR filtering .
# ------------------------------------------------------------------
@pytest.fixture(scope="module")
def generator(rate):
    """Create object to mimic data streaming """
    # Signal of 300 points (sum of two sinus at 0.5 Hz  and 10 Hz) sampled at 50 kHz.

    n = int(rate * 30)  # 30 seconds of data
    m = 3

    f_carrier = 0.5
    a_carrier = 1.0

    f_noise = 10
    a_noise = 0.1

    tm.K = m  # tm will generate 7 columns
    original = tm.makeTimeDataFrame(n, freq="L").rename(
        columns={"A": "noise", "B": "carrier", "C": "signal",}
    )
    original.index = original.index[0] + pd.to_timedelta(np.arange(n) / rate, unit="s")
    t = (original.index - original.index[0]) / np.timedelta64(1, "s")
    original["noise"] = a_noise * np.sin(2 * np.pi * f_noise * t)
    original["carrier"] = a_carrier * np.sin(2 * np.pi * f_carrier * t)
    original["signal"] = original.carrier + original.noise
    data = ReadData(original[["signal"]])
    data._rate = rate
    return data


# -----------------------------------------------------------------------------
# Test that filtering online (chunk by chunk) is the same as filtering offline
# -----------------------------------------------------------------------------
def test_cascade_iirfilter(generator):
    """ Test IIRFilter cascade  """
    rate = generator._rate
    cutoff_hz = 3
    # create filter
    node_iir = IIRFilter(
        rate=rate, frequencies=[cutoff_hz], filter_type="lowpass", order=3
    )

    # Filter online (chunk by chunk)
    # --------------
    # reset the data streamer
    generator.reset()
    looper = Looper(node=node_iir, generator=generator)
    cascade_output, _ = looper.run(chunk_size=5)

    # Filter offline (whole data)
    # --------------
    # reset the data streamer
    generator.reset()
    looper = Looper(node=node_iir, generator=generator)
    continuous_output, _ = looper.run(chunk_size=None)

    # assert filters coeffs are correct
    expected_sos = np.array(
        [
            [0.00475052, 0.00950105, 0.00475052, 1.0, -0.6795993, 0.0],
            [1.0, 1.0, 0.0, 1.0, -1.57048578, 0.68910035],
        ]
    )
    np.testing.assert_array_almost_equal(node_iir._sos, expected_sos)

    # assert signal filtered offline and online are the same after the warmup period.
    order = 3
    warmup = 100 * (order) / (node_iir._rate)
    np.testing.assert_array_almost_equal(
        continuous_output.iloc[int(warmup * node_iir._rate) :].values,
        cascade_output.iloc[int(warmup * node_iir._rate) :].values,
        3,
    )


def test_cascade_firfilter(generator):
    """ Test FIRFilter"""
    rate = generator._rate

    # create the filter
    node_fir = FIRFilter(
        rate=rate, columns="all", order=20, frequencies=[3, 4], filter_type="lowpass"
    )
    expected_coeffs = np.array(
        [
            -0.00217066,
            -0.00208553,
            -0.00108039,
            0.00392436,
            0.01613796,
            0.03711417,
            0.06535715,
            0.09608169,
            0.12241194,
            0.13763991,
            0.13763991,
            0.12241194,
            0.09608169,
            0.06535715,
            0.03711417,
            0.01613796,
            0.00392436,
            -0.00108039,
            -0.00208553,
            -0.00217066,
        ]
    )

    # Filter online (chunk by chunk)
    # --------------
    # reset the data streamer
    generator.reset()
    looper = Looper(node=node_fir, generator=generator)
    cascade_output, cascade_meta = looper.run(chunk_size=5)

    # Filter offline (whole data)
    # --------------
    # reset the data streamer
    generator.reset()
    looper = Looper(node=node_fir, generator=generator)
    cascade_output, metas = looper.run(chunk_size=None)

    # Filter offline
    # --------------
    node_fir.i.data = generator._data.copy()
    node_fir.update()
    continuous_output = node_fir.o.data

    delay = cascade_meta[0]["delay"]

    # assert filters coeffs are correct
    np.testing.assert_array_almost_equal(node_fir._coeffs, expected_coeffs)

    # assert signal filtered offline and online are the same
    warmup = delay * 2
    pd.testing.assert_frame_equal(
        continuous_output.iloc[int(warmup * node_fir._rate) :],
        cascade_output.iloc[int(warmup * node_fir._rate) :],
        check_less_precise=3,
    )

    # correct for induced delay
    fir_o_delayed = cascade_output.copy()
    fir_o_delayed.index -= delay * np.timedelta64(1, "s")


def test_bandpass_power_ratio():
    # Prepare a reproducible example
    np.random.seed(42)
    rate = 512
    n = int(rate * 30)  # 30 seconds of data
    lo, hi = 10, 20
    original = tm.makeTimeDataFrame(n)
    original.index = original.index[0] + pd.to_timedelta(np.arange(n) / rate, unit="s")

    node_iir = IIRFilter(rate=rate, frequencies=[lo, hi], filter_type="bandpass")
    node_fir = FIRFilter(
        rate=rate, frequencies=[lo, hi], filter_type="bandpass", order=rate
    )

    for node in [node_iir, node_fir]:
        node.i.data = original.copy()
        node.update()
        filtered = node.o.data

        freqs = np.fft.fftfreq(n, 1 / rate)

        # Compare in frequency domain
        x = original.values
        X = np.fft.fft(x, axis=0)
        y = filtered.values
        Y = np.fft.fft(y, axis=0)

        f_tol = 1.5  # frequency tolerance from mne
        inner_mask = (freqs > (lo + f_tol)) & (freqs < (hi - f_tol))
        outer_mask = (freqs < (lo - f_tol)) | (freqs > (hi + f_tol))
        pass_ratio = np.abs(Y[inner_mask]) / np.abs(X[inner_mask])
        block_ratio = np.abs(Y[outer_mask]) / np.abs(X[outer_mask])

        # Note for the asserts: absolute tolerances taken from mne would be
        # 0.02 for pass and 0.20 for blocked.
        # We are currently not as good so we are being a bit more relaxed (0.05)
        # for the pass ratio
        np.testing.assert_allclose(np.mean(pass_ratio, axis=0), 1, atol=0.05)
        np.testing.assert_allclose(np.mean(block_ratio, axis=0), 0, atol=0.20)


def test_custom_sos(generator):
    """ Test sos parameter from IIRFilter"""
    rate = generator._rate

    # sos has valid form
    sos = np.array(
        [
            [0.00475052, 0.00950105, 0.00475052, 1.0, -0.6795993, 0.0],
            [1, 1.0, 0.0, 1.0, -1.57048578, 0.68910035],
        ]
    )
    node_iir_custom1 = IIRFilter(
        rate=rate, order=None, frequencies=None, filter_type=None, sos=sos
    )
    generator.reset()
    chunk = generator.next(5).copy()
    node_iir_custom1.i.data = chunk.copy()
    node_iir_custom1.update()

    # sos does not have valid form
    with pytest.raises(ValueError):
        node_iir_custom2 = IIRFilter(
            rate=rate, order=None, frequencies=None, filter_type=None, sos=sos[:, :5]
        )
        node_iir_custom2.i.data = chunk.copy()
        node_iir_custom2.update()



"""This module contains nodes for time index processing"""

import numpy as np

from timeflux.core.node import Node


class DelayIndex(Node):
    def __init__(self, delay=None):
        super().__init__()
        self._delay = None
        self._delay = self.to_timedelta(delay)

    def to_timedelta(self, duration):
        if duration is None:
            return None
        return np.timedelta64(1, "us") * (duration * 1e6)

    def update(self):
        if not self.i.ready():
            return
        if self._delay is None:
            self._delay = self.to_timedelta(self.o.meta.get("delay"))
        if self._delay is not None:
            self.o = self.i
            self.o.data.index -= self._delay



from importlib import import_module


def make_object(fullname, params=None):
    """
    Args:
        fullname (str): full name of the object (ie. '.'.join([o.__module__, o.__name__]))
        params (dict|None): keyword arguments to initialize the object
    Returns:
        Object, instance of the class.
    """
    params = params or {}

    module_name, class_name = fullname.rsplit(".", 1)

    try:
        module = import_module(module_name)
    except ImportError:
        raise ImportError(f"Could not import module {module_name}")
    try:
        object_instance = getattr(module, class_name)(**params)
    except AttributeError:
        raise ValueError(f"Module {module_name} has no object {class_name}")
    return object_instance



import logging

import numpy as np
from scipy import signal

LOGGER = logging.getLogger("timeflux." + __name__)


def _get_com_factor(com, span, halflife, alpha):
    valid_count = sum(1 for _ in filter(None.__ne__, [com, span, halflife, alpha]))
    if valid_count > 1:
        raise ValueError("comass, span, halflife, and alpha " "are mutually exclusive")
    # Convert to smoothing coefficient; domain checks ensure 0 < alpha <= 1
    if com is not None:
        if com < 0:
            raise ValueError("comass must satisfy: com >= 0")
        alpha = 1 / (1 + com)
    elif span is not None:
        if span < 1:
            raise ValueError("span must satisfy: span >= 1")
        alpha = 2 / (span + 1)
    elif halflife is not None:
        if halflife <= 0:
            raise ValueError("halflife must satisfy: halflife > 0")
        alpha = 1 - np.exp(np.log(0.5) / halflife)
    elif alpha is not None:
        if alpha <= 0 or alpha > 1:
            raise ValueError("alpha must satisfy: 0 < alpha <= 1")
    else:
        raise ValueError("Must pass one of com, span, halflife, or alpha")

    return float(alpha)


def _hz_to_nyq(frequencies, nyq):
    if isinstance(frequencies, list):
        return [freq / nyq for freq in frequencies]
    else:
        return frequencies / nyq


def _nyq_to_hz(frequencies, nyq):
    if isinstance(frequencies, list):
        return [freq * nyq for freq in frequencies]
    else:
        return frequencies * nyq


def design_edges(frequencies, nyq, mode):
    """Design filter edges.

    Args:
        frequencies (list): Transition frequencies in Hz.
        mode (str): Filter mode (`lowpass`, `highpass`, `bandstop` or `bandpass`).
        nyq: Nyquist frequency (half sampling rate).

    Returns:
        array: freqs. Updated frequencies with transition bands.
        array: gains.  Filter gain at frequency sampling points.
        list: wp.  Passband edge frequencies.
        list: ws. Stopband edge frequencies.

    **Filter edges design**

    The -6 dB point for all filters is in the middle of the transition band.

    If no transition band is given, default is to use:

        * ``l_trans_bandwidth`` =
          .. math::  min(max(l_{freq} * 0.25, 2), l_{freq})
        * ``h_trans_bandwidth`` =
          .. math:: min(max(h_{freq} * 0.25, 2.), rate / 2. - h_{freq})

    **Band-pass filter**

    The frequency response is (approximately) given by:

    .. image:: /static/image/edges_bandpass.svg

    ..
    ..
    ..   1-|               ----------
    ..     |             /|         | \
    .. \|H\| |            / |         |  \
    ..     |           /  |         |   \
    ..     |          /   |         |    \
    ..   0-|----------    |         |     --------------
    ..     |         |    |         |     |            |
    ..     0        Fs1  Fp1       Fp2   Fs2          Nyq


    Where:

        * l_trans_bandwidth  = Fp1 - Fs1 in Hz
        * Fh_trans_bandwidth = Fs2 - Fp2 +  in Hz
        * ``freqs`` = `[Fp1, Fs1, Fs2, Fp2]`


    **Band-stop filter**

    The frequency response is (approximately) given by:

    .. image:: /static/image/edges_bandstop.svg

    ..
    ..        1-|---------                   ----------
    ..          |         \                 /
    ..      \|H\| |          \               /
    ..          |           \             /
    ..          |            \           /
    ..        0-|             -----------
    ..          |        |    |         |    |        |
    ..          0       Fp1  Fs1       Fs2  Fp2      Nyq


    Where:

        * l_trans_bandwidth  = Fs1 - Fp1 in Hz
        * Fh_trans_bandwidth = Fp2 - Fs2 +  in Hz
        * ``freqs`` = `[Fp1, Fs1, Fs2, Fp2]`




    **Low-pass filter**

        The frequency response is (approximately) given by:

        .. image:: /static/image/edges_lowpass.svg


    ..    1-|------------------------
    ..      |                        \
    ..  \|H\| |                         \
    ..      |                          \
    ..      |                           \
    ..    0-|                            ----------------
    ..      |                       |    |              |
    ..      0                      Fp  Fstop           Nyq

    Where :

        * h_trans_bandwidth = Fstop - Fp in Hz
        * ``freqs`` = `[Fp, Fstop]`

    **High-pass filter**

        The frequency response is (approximately) given by:

    .. image:: /static/image/edges_highpass.svg


    ..
    ..    1-|             -----------------------
    ..      |            /
    ..  \|H\| |           /
    ..      |          /
    ..      |         /
    ..    0-|---------
    ..      |        |    |                     |
    ..      0      Fstop  Fp                   Nyq

    Where :

        * l_trans_bandwidth = Fp - Fstop in Hz
        * ``freqs`` = `[Fstop, Fp]`


    Notes:

        Adapted from mne.filters, see the documentation of:

        * `mne.filters <https://martinos.org/mne/stable/generated/mne.filter.construct_fir_filter.html>`_

    """

    # normalize frequencies
    _normalized_freqs = [f / nyq for f in frequencies]

    if mode == "highpass":
        if len(frequencies) == 1:
            l_freq = frequencies[0]
            l_trans_bandwidth = min(max(l_freq * 0.25, 2), l_freq)
            frequencies = [
                l_freq - l_trans_bandwidth / 2,
                l_freq + l_trans_bandwidth / 2,
            ]
            LOGGER.info(
                "Filter Design: assuming {tb} Hz transition band.".format(
                    tb=frequencies[1] - frequencies[0]
                )
            )
        wp, ws = frequencies[1], frequencies[0]

    elif mode == "lowpass":

        if len(frequencies) == 1:
            h_freq = frequencies[0]
            h_trans_bandwidth = min(max(h_freq * 0.25, 2.0), nyq - h_freq)
            frequencies = [
                h_freq - h_trans_bandwidth / 2,
                h_freq + h_trans_bandwidth / 2,
            ]
            LOGGER.info(
                "Filter design: assuming {tb} Hz transition band.".format(
                    tb=frequencies[1] - frequencies[0]
                )
            )
        wp, ws = _hz_to_nyq(frequencies, nyq)

    elif mode == "bandpass":
        if len(frequencies) == 2:
            lofreq, hifreq = frequencies
            minfreq = lofreq - 1 if lofreq > 1 else lofreq / 2
            maxfreq = hifreq + 1 if hifreq < nyq - 1 else (hifreq + nyq) / 2
            frequencies = [minfreq, lofreq, hifreq, maxfreq]
            LOGGER.info(
                "Filter Design: assuming {tb1} and {tb2} Hz transition band.".format(
                    tb1=frequencies[2] - frequencies[1],
                    tb2=frequencies[3] - frequencies[0],
                )
            )
        wp, ws = (
            _hz_to_nyq([frequencies[1], frequencies[2]], nyq),
            _hz_to_nyq([frequencies[0], frequencies[3]], nyq),
        )

    elif mode == "bandstop":
        if len(frequencies) == 2:
            l_freq, h_freq = frequencies
            l_trans_bandwidth = min(max(l_freq * 0.25, 2), l_freq)
            h_trans_bandwidth = min(max(h_freq * 0.25, 2.0), nyq - h_freq)

            frequencies = [
                l_freq - l_trans_bandwidth / 2,
                l_freq + l_trans_bandwidth / 2,
                h_freq - h_trans_bandwidth / 2,
                h_freq + h_trans_bandwidth / 2,
            ]

            LOGGER.info(
                "Filter Design: assuming {tb1} and {tb2} Hz transition band.".format(
                    tb1=frequencies[3] - frequencies[0],
                    tb2=frequencies[1] - frequencies[2],
                )
            )
        wp, ws = (
            _hz_to_nyq([frequencies[0], frequencies[3]], nyq),
            _hz_to_nyq([frequencies[1], frequencies[2]], nyq),
        )

    else:
        raise ValueError("Unknown filter mode given: {mode} ".format(mode=mode))

    if mode == "bandpass":
        frequencies = [0] + frequencies + [nyq]
        gains = [0, 0, 1, 1, 0, 0]
    elif mode == "bandstop":
        frequencies = [0] + frequencies + [nyq]
        gains = [1, 1, 0, 0, 1, 1]
    elif mode == "highpass":
        frequencies = [0] + frequencies + [nyq]
        gains = [0, 0, 1, 1]
    elif mode == "lowpass":
        frequencies = [0] + frequencies + [nyq]
        gains = [1, 1, 0, 0]
    else:
        raise ValueError("unsupported filter mode: {mode}".format(mode=mode))

    if any([f > nyq for f in frequencies]):
        raise ValueError(
            "One of the given frequencies exceeds the "
            "Nyquist frequency of the signal (%.1f)." % nyq
        )

    return frequencies, gains, wp, ws


def _filter_attenuation(h, frequencies, gains):
    """Compute minimum attenuation at stop frequency.

     Args:
        h (array): Filter coefficients.
        frequencies (list): Transition frequencies normalized.
        gains (array):  Filter gain at frequency sampling points.

     Returns:
        att_db: Minimum attenuation per frequency
        att_freq: Frequencies

    Notes:

        Adapted from mne.filters

    """
    frequencies = np.array(frequencies)
    from scipy.signal import freqz

    _, filt_resp = freqz(h.ravel(), worN=np.pi * frequencies)
    filt_resp = np.abs(filt_resp)  # use amplitude response
    filt_resp[np.where(gains == 1)] = 0
    idx = np.argmax(filt_resp)
    att_db = -20 * np.log10(np.maximum(filt_resp[idx], 1e-20))
    att_freq = frequencies[idx]
    return att_db, att_freq


def construct_fir_filter(rate, frequencies, gains, order, phase, window, design):
    """Construct coeffs of FIR filter.

    Args:
        rate (float): Nominal sampling rate of the input data.
        order (int): Filter order
        frequencies (list): Transition frequencies in Hz.
        design (str|'firwin2'): Design of the transfert function of the filter.
        phase (str|`linear`): Phase response ("zero", "zero-double" or "minimum").
        window (float|`hamming`): The window to use in FIR design, ("hamming", "hann", or "blackman").

    Returns:
        array h. FIR coeffs.

    Notes:

        Adapted from mne.filters, see the documentation of:

        * `mne.filters <https://martinos.org/mne/stable/generated/mne.filter.construct_fir_filter.html>`_
        * `scipy.signal.firwin2 <https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.signal.firwin2.html>`_



    """
    nyq = rate / 2.0
    if design == "firwin2":
        from scipy.signal import firwin2 as design
    else:
        # not implemented yet
        raise ValueError("firwin, remez and firls have not been implemented yet ")

    # issue a warning if attenuation is less than this
    min_att_db = 12 if phase == "minimum" else 20

    if frequencies[0] != 0 or frequencies[-1] != nyq:
        raise ValueError(
            "freq must start at 0 and end an Nyquist (%s), got %s" % (nyq, frequencies)
        )
    gains = np.array(gains)

    if window == "kaiser":
        diffs = np.diff(frequencies)
        width = min(diffs[diffs > 0])
        beta = signal.kaiser_beta(signal.kaiser_atten(order, width / nyq))
        window = ("kaiser", beta)

    # check zero phase length
    N = int(order)
    if N % 2 == 0:
        if phase == "zero":
            LOGGER.info('filter_length must be odd if phase="zero", ' "got %s" % N)
            N += 1
        elif phase == "zero-double" and gains[-1] == 1:
            N += 1
    # construct symmetric (linear phase) filter
    if phase == "minimum":
        h = design(N * 2 - 1, frequencies, gains, fs=rate, window=window)
        h = signal.minimum_phase(h)
    else:
        h = design(N, frequencies, gains, fs=rate, window=window)
    assert h.size == N
    att_db, att_freq = _filter_attenuation(h, frequencies, gains)
    if phase == "zero-double":
        att_db += 6
    if att_db < min_att_db:
        att_freq *= rate / 2.0
        LOGGER.info(
            "Attenuation at stop frequency %0.1fHz is only %0.1fdB. "
            "Increase filter_length for higher attenuation." % (att_freq, att_db)
        )
    return h


def construct_iir_filter(
    rate,
    frequencies,
    filter_type,
    order=None,
    design="butter",
    pass_loss=3.0,
    stop_atten=50.0,
    output="sos",
):
    """Calculate an IIR filter kernel for a given sampling rate.

    Args:
        rate (float): Nominal sampling rate of the input data.
        order (int): Filter order
        frequencies (list): Transition frequencies in Hz.
        filter_type (str): Filter mode (`lowpass`, `highpass`, `bandstop` or `bandpass`).
        design (str|'butter'): Design of the transfert function of the filter (`butter`, `cheby1`, `cheby2`, `ellip`, `bessel`)
        pass_loss (float|`3.0`): For Chebyshev and elliptic filters, provides the maximum ripple in the passband. (dB).
        stop_atten (float|`50.0`): For Chebyshev and elliptic filters, provides the minimum attenuation in the stop band. (dB)

    Returns:
        array: sos. IIR coeffs.

    Notes:

        Adapted from mne.filters, see the documentation of:

        * `mne.filters <https://martinos.org/mne/stable/generated/mne.filter.construct_fir_filter.html>`_
        * `scipy.signal.iirfilter <https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.signal.iirfilter.html>`_
        * `scipy.signal.iirdesign <https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.signal.iirdesign.html>`_


    """

    _minorder = {
        "butter": signal.buttord,
        "cheby1": signal.cheb1ord,
        "cheby2": signal.cheb2ord,
        "ellip": signal.ellip,
    }

    nyq = rate / 2.0

    if order:
        # use order-based design
        wn = _hz_to_nyq(frequencies, nyq)
        out = signal.iirfilter(
            N=order,
            Wn=wn,
            rp=pass_loss,
            rs=stop_atten,
            btype=filter_type,
            ftype=design,
            output=output,
        )
        return out, frequencies
    else:
        frequencies, gains, wp, ws = design_edges(frequencies, nyq, filter_type)
        out = signal.iirdesign(
            wp=wp, ws=ws, gstop=stop_atten, gpass=pass_loss, ftype=design, output=output
        )
        return out, frequencies



"""Tests for detection peaks nodes"""
from numbers import Number

import numpy as np
import os
import pandas as pd
import pandas.util.testing as tm
import pytest
from timeflux.helpers.testing import ReadData, Looper

from timeflux_dsp.nodes.peaks import LocalDetect, RollingDetect


@pytest.fixture(scope="module")
def ppg_generator():
    """Create object to mimic data streaming """
    # Signal of 300 points (sum of two sinus at 0.5 Hz  and 10 Hz) sampled at 50 kHz.

    df = pd.read_csv(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data', 'test_data_ppg.csv'), index_col=None)
    df = pd.DataFrame(
        index=pd.to_datetime(df["index"].values), data=df["PPG"].values, columns=["PPG"]
    )
    ppg_generator = ReadData(df[:200])
    return ppg_generator


def assert_dict_almost_equal(dict1, dict2, decimal=4, keys=None):
    if keys is None:
        np.testing.assert_array_equal(list(dict1.keys()), list(dict2.keys()))
    for k in keys or dict1:
        if isinstance(dict1[k], Number):
            np.testing.assert_almost_equal(dict1[k], dict2[k], decimal=decimal)


def _gen_gaussians(center_locs, sigmas, total_length):
    xdata = np.arange(0, total_length).astype(float)
    out_data = np.zeros(total_length, dtype=float)
    for ind, sigma in enumerate(sigmas):
        tmp = (xdata - center_locs[ind]) / sigma
        out_data += np.exp(-(tmp ** 2))
    return out_data


def test_localdetect_on_gaussians():
    """ Test on a sum of two gaussians if the peak center is well estimated """
    rate = 128
    num_points = 10 * rate
    tm.K = 1  # tm will generate 1 column
    original = tm.makeTimeDataFrame(num_points, freq="L")
    original.index = original.index[0] + pd.to_timedelta(
        np.arange(num_points) / rate, unit="s"
    )

    sigmas = [5.0, 1.0]
    lags = [5.0 / rate, 1.0 / rate]
    center_locs = [1 * rate, 5 * rate]
    intervals = [1.0, 5.0 - 1.0]

    gaussians_values = _gen_gaussians(center_locs, sigmas, num_points)
    peak_times = original.index[center_locs]

    original["A"] = gaussians_values

    data_gaussians = ReadData(original)
    node = LocalDetect(delta=0.5, tol=0.5)
    # loop across chunks
    looper = Looper(data_gaussians, node)
    cascade_output, _ = looper.run(chunk_size=5)
    estimation_times = [
        pd.Timestamp(event["extremum_time"])
        for event in cascade_output[cascade_output.label == "peak"].data.values
    ]

    assert estimation_times == list(peak_times)

    data_peaks = pd.concat(
        [
            pd.DataFrame(meta, index=[pd.Timestamp(meta["extremum_time"])])
            for meta in cascade_output[cascade_output.label == "peak"].data.values
        ],
        ignore_index=True,
    )

    expected_peaks = pd.DataFrame(
        dict(column_name=["A", "A"], value=[1.0, 1.0], interval=intervals, lag=lags)
    )
    pd.testing.assert_frame_equal(
        data_peaks.drop(["now", "extremum_time", "detection_time"], axis=1),
        expected_peaks,
        check_like=True,
        atol=1e-4
    )


def test_localdetect_on_ppg(ppg_generator):
    node = LocalDetect(delta=0.5, tol=0.5)
    # reset generator
    ppg_generator.reset()
    # loop across chunks
    looper = Looper(ppg_generator, node)
    cascade_output, _ = looper.run(chunk_size=5)

    expected_extremum_times = pd.DatetimeIndex(
        [
            "2018-11-19 11:06:39.62",
            "2018-11-19 11:06:39.79",
            "2018-11-19 11:06:40.61",
            "2018-11-19 11:06:40.76",
            "2018-11-19 11:06:41.56",
            "2018-11-19 11:06:41.71",
        ]
    )
    actual_extremum_times = pd.DatetimeIndex(
        [data["extremum_time"] for data in cascade_output.data.values]
    )
    expected_labels = ["peak", "valley"] * 3

    expected_data_peak_2 = {
        "value": np.array([0.95]),
        "lag": 0.048,
        "interval": 0.98,
        "column_name": "PPG",
    }
    expected_data_valley_3 = {
        "value": np.array([-1.05]),
        "lag": 0.48,
        "interval": 0.96,
        "column_name": "PPG",
    }

    pd.testing.assert_index_equal(
        expected_extremum_times, actual_extremum_times.round("10ms")
    )
    np.testing.assert_array_equal(expected_labels, cascade_output.label.values)
    assert_dict_almost_equal(
        expected_data_peak_2,
        cascade_output.data.values[2],
        decimal=2,
        keys=["value", "lag", "interval", "column_name"],
    )
    assert_dict_almost_equal(
        expected_data_valley_3,
        cascade_output.data.values[3],
        decimal=2,
        keys=["value", "lag", "interval", "column_name"],
    )


def test_rollingdetect(ppg_generator):
    node = RollingDetect(length=0.8, tol=0.5, rate=64)

    # reset generator
    ppg_generator.reset()
    # loop across chunks
    looper = Looper(ppg_generator, node)
    cascade_output, _ = looper.run(chunk_size=5)

    expected_index = pd.DatetimeIndex(
        [
            "2018-11-19 11:06:39.62",
            "2018-11-19 11:06:39.79",
            "2018-11-19 11:06:40.61",
            "2018-11-19 11:06:40.76",
        ]
    )

    expected_labels = ["peak", "valley", "peak", "valley"]
    expected_data_peak_2 = {
        "value": np.array([0.95]),
        "lag": 0.84,
        "interval": 0.98,
        "column_name": "PPG",
    }
    expected_data_valley_3 = {
        "value": np.array([-1.05]),
        "lag": 0.84,
        "interval": 0.96,
        "column_name": "PPG",
    }

    pd.testing.assert_index_equal(expected_index, cascade_output.index.round("10ms"))
    np.testing.assert_array_equal(expected_labels, cascade_output.label.values)
    assert_dict_almost_equal(
        expected_data_peak_2,
        cascade_output.data.values[2],
        decimal=2,
        keys=["value", "lag", "interval", "column_name"],
    )
    assert_dict_almost_equal(
        expected_data_valley_3,
        cascade_output.data.values[3],
        decimal=2,
        keys=["value", "lag", "interval", "column_name"],
    )



"""Tests for decimation nodes"""
import numpy as np
import pandas as pd
import pytest
from timeflux.helpers.testing import DummyData, Looper

from timeflux_dsp.nodes.filters import DropRows, Resample


@pytest.fixture(scope="module")
def generator(rate=10, jitter=0.05):
    """Create object to mimic data streaming """
    generator = DummyData(rate=rate, jitter=jitter)
    return generator


def test_droprows_1(generator):
    """ Test DropRows
    Test that the node correctly maintains an internal state, when size of
    chunk is not a multiple of factor
    """
    factor = 2
    generator.reset()
    node = DropRows(factor=factor, method=None)
    node.i.data = generator.next(5)
    node.update()
    expected_o = generator._data.iloc[[1, 3]]
    expected_p = generator._data.iloc[[4]]
    pd.testing.assert_frame_equal(node.o.data, expected_o)
    pd.testing.assert_frame_equal(node._previous, expected_p)


def test_droprows_2(generator):
    """ Test DropRows
    test for factor = 2, 3, 4, 8 that the output is equivalent to applying a
    rolling window and taking the mean over the samples.
    size of chunk is 5 rows.
    """

    for factor in [2, 3, 4, 8]:
        generator.reset()
        node = DropRows(factor=factor, method="mean")
        looper = Looper(node=node, generator=generator)
        out_data, _ = looper.run(chunk_size=10)
        expected = (
            generator._data.rolling(window=factor, min_periods=factor, center=False)
            .mean()
            .iloc[np.arange(factor - 1, len(generator._data), factor)]
        )
        pd.testing.assert_frame_equal(
            out_data.iloc[: len(generator._data) // factor], expected
        )


def test_resample(generator):
    """ Test Resample
    """
    factor = 2

    generator.reset()
    node = Resample(factor=factor)
    chunk = generator.next(5)

    expected_o = pd.DataFrame(
        data=np.array(
            [
                [0.47955525, 0.37772175, 0.77298, 0.830601, 0.55936975],
                [0.51563125, 0.82348875, 0.844093, 0.137803, 0.45838375],
            ]
        ),
        index=generator._data.index[[0, 2]],
    )
    expected_p = generator._data.iloc[[4]]

    node.i.data = chunk
    node.update()
    pd.testing.assert_frame_equal(node.o.data, expected_o)
    pd.testing.assert_frame_equal(node._previous, expected_p)
